"""
Python wrapper for the WhatsApp Cloud API.
    - GitHub: https://github.com/david-lev/pywa
    - Docs: https://pywa.readthedocs.io/
    - PyPI: https://pypi.org/project/pywa/
"""

from pywa.client import WhatsApp

__version__ = "1.13.0-rc.1"
__author__ = "David Lev"
__license__ = "MIT"
